"""Randomly picks n samples and injects them into the prompt."""
import random
from .selector import Selector
random.seed(42)


class RandomSampler(Selector):
    """Randomly picks n samples and injects them into the prompt."""

    def select_examples(self):
        """Select examples for the task randomly""" 
        random.shuffle(self.all_examples)
        self.selected_examples = self.all_examples[:self.num_samples]
        self.task.selected_examples = self.selected_examples
        return self

    def __repr__(self):
        return f"""RandomSampler(task={self.task},
        num_samples={self.num_samples}"""



"""A class for a Generic Prompt."""
from pydantic import Field, BaseModel
from prompt_learner.optimizers.selectors.selector import Selector
from prompt_learner.optimizers.selectors.random_sampler import RandomSampler
from prompt_learner.templates.template import Template


class Prompt(BaseModel):
    """Defines the contract for a Generic Prompt."""
    template: Template
    selector: Selector = Field(description="Selector for the task.")
    prompt: str = Field(description="Final prompt string.", default="")

    def select_examples(self):
        """Select examples for the task."""
    
    def assemble_prompt(self):
        """Assemble the prompt."""
        self.prompt = f"""{self.template.descriptor}{self.template.examples_preamble}
        {self.template.format_examples(self.selector.selected_examples)}"""

    def add_inference(self, text: str, context: str = ""):
        """Add inference sample"""
        self.prompt = self.prompt + self.template.add_prediction_sample(text,context)


"""This module contains the OpenAI class
which is used to interact with the OpenAI API."""

import os
from langchain_openai import ChatOpenAI
from dotenv import load_dotenv
from .adapter import Adapter


class OpenAI(Adapter):
    """An adapter for an OpenAI language model call"""
    def __init__(self, temperature: float = 0.0, max_tokens: int = 1024):
        super().__init__(temperature, max_tokens)
        load_dotenv()
        self.llm = ChatOpenAI(
            openai_api_key=os.getenv('OPENAI_API_KEY'),
            temperature=self.temperature,
            max_tokens=self.max_tokens)



"""Defines the contract for a text to sql task."""

from prompt_learner.examples.example import Example
from .task import Task


class SQLGenerationTask(Task):
    """SQLGeneration"""

    def validate_example(self, example: Example):
        """Validate the example for SQL Generation task."""
        if example.context is None: #need schema as context
            return False
        return True



"""This module contains the Anthropic class,
which is an adapter for the Anthropic language model API."""

import os
import re
from langchain_anthropic import ChatAnthropic
from dotenv import load_dotenv
from .adapter import Adapter


class Anthropic(Adapter):
    """An adapter for an Anthropic language model call"""
    def __init__(self, temperature: float = 0.0, max_tokens: int = 1024):
        super().__init__(temperature, max_tokens)
        load_dotenv()
        self.llm = ChatAnthropic(
                anthropic_api_key=os.getenv('ANTHROPIC_API_KEY'),
                model="claude-3-haiku-20240307",
                temperature=self.temperature,
                max_tokens=self.max_tokens)

    def extract_xml_tag(self, data: str, tag: str) -> str:
        """Extracts the data between the XML tags."""
        open_tag = "<" + tag + ">"
        close_tag = "</" + tag + ">"
        try:
            data = (data.split(open_tag)[1]).split(close_tag)[0].strip()
        except IndexError:
            pass
        data = re.sub(r"^\\n|\\n$", "", data)
        return data
    
    def process_output(self, output: str):
        """Process the output from the language model."""
        return self.extract_xml_tag(output.content.strip(), "label")



"""Compute accuracy metric for the prompt+model."""
from copy import deepcopy
from prompt_learner.tasks.task import Task
from prompt_learner.prompts.prompt import Prompt
from prompt_learner.adapters.adapter import Adapter

class Accuracy:
    """Defines the contract for the Accuracy Metric."""
    def __init__(self, task: Task):
        self.task = task

    def compute(self, prompt: Prompt, adapter: Adapter, test=False) -> tuple[float, int]:
        """Compute the accuracy of the model."""
        correct = 0
        total = 0
        if test is False:
            examples = self.task.examples
        else:
            examples = self.task.test_examples
        for example in examples:
            if example not in self.task.selected_examples:
                total += 1
                temp_prompt = deepcopy(prompt)
                temp_prompt.add_inference(example.text)
                prediction = self.task.predict(adapter, temp_prompt.prompt)
                prediction = prediction.strip()
                prediction = prediction.strip('\n')
                if example.label == prediction:
                    correct += 1
                else:
                    print("Incorrect prediction:", example.text, example.label,prediction)
        return (correct / total, total)



"""This module contains the AnthropicCompletionTemplate class"""
from typing import List
from prompt_learner.examples.example import Example
from .template import Template


class AnthropicCompletionTemplate(Template):
    """This class generates a template for Anthropic completions"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        tasks_with_labels = ["Classification", "Tagging"]
        self.descriptor = f"""You are a helpful AI assistant.\nYou are helping a user with a {self.task_type} task.\nThe user gives you the following task description.\n<task_description>{self.task_description}</task_description>\n"""
        if self.allowed_labels:
            self.descriptor += """You have to select from the following labels.\n<allowed_labels>{self.allowed_labels}</allowed_labels>"""
        if self.task_type in tasks_with_labels:
            self.prediction_preamble = f"""Given the text, you have to now predict the labels from list of allowed labels - {self.allowed_labels}.\nOutput only the label(s) and close the <label> tag."""
        elif self.task_type == "SQLGeneration":
            self.prediction_preamble = """Given the text, you have to now generate a SQL query.\nOnly the SQL query is expected."""
        else:  #generic preamble for prediction
            self.prediction_preamble = """Given the text, you have to now predict."""
        self.examples_preamble = """Here are a few examples to help you understand the task better.\n"""
    
    def format_examples(self, examples: List[Example]):
        """Formats the task examples into a string."""
        tasks_with_labels = ["Classification", "Tagging"]
        examples_str = ""
        for example in examples:
            if self.task_type in tasks_with_labels:
                examples_str += f"""<example>\n<text> {example.text}</text>\n<label> {example.label}</label>\n</example>\n"""
            elif self.task_type == "SQLGeneration":
                examples_str += f"""<example>\n<schema> {example.context}</schema>\n<text> {example.text}</text>\n<SQL> {example.label}</SQL>\n</example>\n"""
            else:
                examples_str += f"""<example>\n<text> {example.text}</text>\n<output> {example.label}</output>\n</example>\n"""
        return examples_str
    
    def add_prediction_sample(self, text: str, context: str):
        """Add prediction sample to task."""
        tasks_with_labels = ["Classification", "Tagging"]
        prediction_preamble = self.prediction_preamble + f"""\n <text> {text} <text>"""
        if self.task_type in tasks_with_labels:
            return prediction_preamble + "<label>"
        elif self.task_type == "SQLGeneration":
            return prediction_preamble + f"""\n <schema>{context} <\schema>\m <SQL>"""
        else:
            return prediction_preamble + "<output>"



"""Generic Task class."""

from typing import List, Optional
from pydantic import BaseModel, Field
from prompt_learner.examples.example import Example
from prompt_learner.adapters.adapter import Adapter


class Task(BaseModel): #TO_DO enum for all task types
    """Defines the contract for a Generic task."""
    description: str = Field(description="The name of the task.")
    allowed_labels: List[str] = Field(description="Allowed labels for task.", default=[])
    examples: List[Example] = []
    selected_examples: List[Example] = []
    test_examples: List[Example] = []

    def validate_example(self, example: Example):
        """Validate the example for the task."""
        # This method will be overridden in subclasses

    def add_example(self, example: Example, test=False):
        """Add an example to the task."""
        if not self.validate_example(example):
            raise ValueError(f"""Label '{example.label}' is not in
                             allowed labels for this task.""")
        if test is False:
            self.examples.append(example)
            self.selected_examples.append(example)
        else:
            self.test_examples.append(example)

    def predict(self, adapter: Adapter, prompt: str):
        """Predict the label for the given text using the prompt."""
        return adapter.process_output(adapter.llm.invoke(prompt))
    
    def validate_prediction(self, text: str, prediction: str):
        """Validate the prediction."""
        example = Example(text=text, label=prediction)
        return self.validate_example(example)



"""A Generic Example Selector class."""
from typing import List
from pydantic import BaseModel, Field
from prompt_learner.tasks.task import Task
from prompt_learner.examples.example import Example


class Selector(BaseModel):
    """Defines the contract for a Generic Example Selector."""
    all_examples: List[Example] = []
    selected_examples: List[Example] = []
    task: Task
    num_samples: int = Field(description="Number of examples to samples", default=1)
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.all_examples = self.task.examples

    def select_examples(self):
        """Select examples for the task."""
        self.all_examples = self.task.examples
        self.selected_examples = self.examples
        self.task.selected_examples = self.select_examples
        # This method will be overridden in subclasses



"""Defines the contract for a classification task."""

from typing import List
from prompt_learner.examples.example import Example
from .task import Task


class ClassificationTask(Task):
    """Classification"""
    allowed_labels: List[str] = []

    def validate_example(self, example: Example):
        """Validate the example for tagging task."""
        labels = example.label.split(',')  # Handle multiple labels
        allowed_labels_set = set(self.allowed_labels)  # Convert to set
        if not all(label.strip() in allowed_labels_set for label in labels):
            return False
        return True



"""Class for the CoT prompting."""
from pydantic import Field
from .prompt import Prompt


class CoT(Prompt):
    """Think step by step"""
    custom_intructions: str = Field(description="""Custom intructions for Chain
                                 of Thought Prompting""",
                                 default="Think step by step.\n")
    
    def assemble_prompt(self):
        """Assemble the prompt."""
        self.prompt = f"""{self.template.descriptor}{self.template.examples_preamble}{self.template.format_examples(self.selector.selected_examples)}{self.custom_intructions}"""



"""Randomly picks n samples and injects them into the prompt."""
import random
from .selector import Selector
random.seed(42)


class StratifiedSampler(Selector):
    """Randomly picks n samples for each label.
    Total number of samples selected = num_samples * num_labels"""

    def select_examples(self):
        """Select n examples from each class for the task randomly"""
        self.all_examples = self.task.examples
        self.selected_examples = []
        for label in self.task.allowed_labels:
            examples = [example for example in self.all_examples if example.label == label]
            random.shuffle(examples)
            self.selected_examples += examples[:self.num_samples]
        self.task.selected_examples = self.selected_examples

    def __repr__(self):
        return f"""StratifiedSampler(task={self.task},
        num_samples={self.num_samples}"""



"""Picks n most diverse samples using tf-idf as a measure."""
import random
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from .selector import Selector
random.seed(42)


class DiverseSampler(Selector):
    """Picks n most diverse samples."""

    def select_examples(self):
        """Select n examples based on tf-idf diversity."""
        self.all_examples = self.task.examples
        self.selected_examples = []
        texts = [example.text for example in self.all_examples]

        # Vectorize text data
        vectorizer = TfidfVectorizer()
        tfidf_matrix = vectorizer.fit_transform(texts)

        # Compute pairwise cosine similarity matrix
        sim_matrix = cosine_similarity(tfidf_matrix)

        # Find the indices of the n most diverse samples
        indices = []
        while len(indices) < self.num_samples:
            if not indices:
                # Start with the first text if no indices are selected yet
                current_index = 0
            else:
                # Select the next text that has the minimum maximum similarity
                # to already selected texts
                current_index = np.argmin(np.max(sim_matrix[:, indices], axis=1))
            
            # Add the index of the selected text to the list
            indices.append(current_index)

        self.selected_examples = [self.all_examples[i] for i in indices]
        self.task.selected_examples = self.selected_examples



"""This module contains the OpenAICompletionTemplate class"""
from typing import List
from prompt_learner.examples.example import Example
from .template import Template


class OpenAICompletionTemplate(Template):
    """This class generates a template for OpenAI completions"""
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        tasks_with_labels = ["Classification", "Tagging"]
        self.descriptor = f"""You are a helpful AI assistant.\nYou are helping a user with a {self.task_type} task.\nThe user gives you the following task description.\n{self.task_description}\n"""
        if self.allowed_labels:
            self.descriptor += f"""You have to select from the following labels.\n{self.allowed_labels}."""
        if self.task_type in tasks_with_labels:
            self.prediction_preamble = f"""Given the text, you have to now predict the labels from the list of allowed labels - {self.allowed_labels}."""
        elif self.task_type == "SQLGeneration":
            self.prediction_preamble = """Given the text, you have to now generate a SQL query."""
        else:  #generic preamble for prediction
            self.prediction_preamble = """Given the text, you have to now predict."""
        self.examples_preamble = """Here are a few examples to help you understand the task better.\n"""
       
    def format_examples(self, examples: List[Example]):
        """Formats the task examples into a string."""
        tasks_with_labels = ["Classification", "Tagging"]
        examples_str = ""
        for example in examples:
            if self.task_type in tasks_with_labels:
                examples_str += f"""text: {example.text}\nlabel: {example.label}\n"""
            elif self.task_type == "SQLGeneration":
                examples_str += f"""schema: {example.context}\ntext: {example.text}\nSQL: {example.label}\n"""
            else: #generic example format
                examples_str += f"""
                text: {example.text}\noutput: {example.label}\n"""
        return examples_str

    def add_prediction_sample(self, text: str, context: str = None):
        """Add prediction sample to task."""
        tasks_with_labels = ["Classification", "Tagging"]
        if self.task_type in tasks_with_labels:
            return self.prediction_preamble + f"""\ntext: {text}"""+ "\nlabel:"
        elif self.task_type == "SQLGeneration":
            return self.prediction_preamble + f"""\nschema: {context}\ntext: {text}\nSQL: """
        else:
            return self.prediction_preamble + "\noutput:"



"""Class for a Generic Adapter"""


class Adapter:
    """Defines the contract for a Generic Adapter."""
    def __init__(self, temperature: float = 1.0, max_tokens: int = 1024):
        self.temperature = temperature
        self.max_tokens = max_tokens

    def process_output(self, output: str):
        """Process the output from the language model."""
        return output.content.strip()


"""Class for a Generic prompt Template"""

from typing import List, Any
from pydantic import BaseModel, Field
from prompt_learner.tasks.task import Task
from prompt_learner.examples.example import Example


class Template(BaseModel):
    """Defines the contract for a Generic template."""
    task_description: str = Field(default="", description="Describes the task")
    task_type: str = Field(default="", description="Type of task")
    task: Task
    allowed_labels: List[Any] = Field(default=[],
                                      description="Allowed labels for task")
    examples: List[Any] = Field(default=[],
                                description="Examples for the task")
    descriptor: str = Field(default="", description="Descriptor for the task")
    examples_preamble: str = Field(default="",
                                   description="Preamble for examples")
    prediction_preamble: str = Field(default="",
                                     description="Preamble forprediction")

    def __init__(self, **kwargs):
    
        super().__init__(**kwargs)
        self.task_description = self.task.description
        self.task_type = self.task.__doc__
        self.allowed_labels = self.task.allowed_labels
        
    def format_examples(self, examples: List[Example]):
        """Add an example to the task."""
        # This method will be overridden in subclasses

    def add_prediction_sample(self, text: str, context: str):
        """Add inference instructions to task."""
        # This method will be overridden in subclasses



"""Example Pydantic model for a Task"""
from pydantic import BaseModel, Field


class Example(BaseModel):
    """An Example for a Task"""
    text: str = Field(description="Input text")
    context: str = Field(description="Context for the input text", default="")
    label: str = Field(description="Output label")
